"""Module that defines the backend for svg rendering to a pygame surface. This relies on the cairosvg package."""

from lxml import etree as ET
import numpy as np
import cairosvg
import pygame
import re


class CairoSVGSurface:
    """A surface implementation that is able to render svg graphics to a pygame window."""

    def __init__(self, surface_size: tuple[int, int]):
        """Constructor.

        Args:
            surface_size (tuple[int,int]) : size of the render surface.
        """
        super().__init__()
        self._svg_source = """<svg xmlns="http://www.w3.org/2000/svg"></svg>"""
        self._svg_tree = ET.fromstring(self._svg_source)
        self._surface = pygame.Surface(surface_size)
        # position defined in the svg, this is NOT in pixels but in svg coordinates
        self._svg_position = (0, 0)
        # size defined in the svg, this is NOT in pixels but in svg coordinates
        self._svg_size = (0, 0)
        # the scaling factor (maintaining aspect ratio) of the svg to ensure it fits in the surface
        self._scaling_factor = 1.0
        # pixel offset when rendering the svg to the surface, this is needed because
        # the svg is automatically centered in the surface during the render
        self._surface_offset = (0, 0)
        # size of the pygame window (or surface) we are rendering to
        self._window_size = (0, 0)

    def update(self, svg_tree: ET.ElementBase):
        """Update internal svg data.

        Args:
            svg_tree (ET.ElementBase): root of the svg tree
        """
        self._svg_tree = svg_tree
        # svg element must be at the root
        assert self._svg_tree.tag.endswith("svg")
        # the svg size MUST be defined...
        self._svg_size = (
            int(self._svg_tree.get("width")),
            int(self._svg_tree.get("height")),
        )
        self._svg_position = (
            int(self._svg_tree.get("x", 0)),
            int(self._svg_tree.get("y", 0)),
        )
        # convert the svg tree to its source string so that Cairo can render it
        self._svg_source = ET.tostring(
            self._svg_tree, method="c14n2", with_comments=False
        )

    @property
    def svg_size(self) -> tuple[float, float]:
        """Getter for the size of the root svg element (svg coordinates).

        Returns:
            tuple[float,float]: root svg element size
        """
        return self._svg_size

    @property
    def surface_position(self) -> tuple[float, float]:
        """Getter for the position of the surface in the window, this is used to place the surface in the window.

        Returns:
            tuple[float, float]: surface position in the window.
        """
        # compute the surface position based on svg position and scaling factor
        p = self._svg_position
        surface_position = (p[0] * self.scaling_factor, p[1] * self.scaling_factor)
        # center the surface in the window
        surface_size = self.surface_size
        window_size = self._window_size
        centering_offset = (
            (window_size[0] - surface_size[0]) / 2,
            (window_size[1] - surface_size[1]) / 2,
        )
        return (
            centering_offset[0] + surface_position[0],
            centering_offset[1] + surface_position[1],
        )

    @property
    def surface_size(self) -> tuple[int, int]:
        """Getter for the size of the render surface (in pixels).

        Returns:
            tuple[int, int]: size of the render surface.
        """
        return self._surface.get_size()

    @property
    def scaling_factor(self) -> float:
        """Getter for the scaling factor (maintaining aspect ratio) of the svg to ensure it fits in the render surface.

        Returns:
            float: scaling factor
        """
        return self._scaling_factor

    def pixel_to_svg(self, point: tuple[float, float]) -> tuple[float, float]:
        """Transforms a point from pixel space to svg space.

        Args:
            point (tuple[float, float]): to transform

        Returns:
            tuple[float, float]: transformed point
        """
        spos = self.surface_position
        sfac = self.scaling_factor
        return (
            (point[0] - self._surface_offset[0] - spos[0]) / sfac,
            (point[1] - self._surface_offset[1] - spos[1]) / sfac,
        )

    def pixel_scale_to_svg_scale(
        self, point: tuple[float, float]
    ) -> tuple[float, float]:
        """Scale a point (typically a (width, height) pair or similar) that lies in pixel space to svg space. Only scaling is applied.

        Args:
            point (tuple[float, float]): point to scale

        Returns:
            tuple[float, float]: scaled point
        """
        sfac = self.scaling_factor
        return (point[0] * sfac, point[1] * sfac)

    def svg_to_pixel(self, point: tuple[float, float]) -> tuple[float, float]:
        """Transform a point from svg space to pixel space.

        Args:
            point (tuple[float, float]): to transform

        Returns:
            tuple[float, float]: transformed point
        """
        raise NotImplementedError()  # TODO

    def svg_scale_to_pixel_scale(
        self, point: tuple[float, float]
    ) -> tuple[float, float]:
        """Scale a point (typically a (width, height) pair or similar) that lies in svg space to pixel space. Only scaling is applied.

        Args:
            point (tuple[float, float]): point to scale

        Returns:
            tuple[float, float]: scaled point
        """
        raise NotImplementedError()  # TODO

    def render(self, window: pygame.Surface, background_color="#ffffff"):
        """Render the svg to the pygame surface `window`.

        Args:
            window (pygame.Surface): pygame surface
            background_color (str, optional): background color. Defaults to white.
        """
        # center the svg in the window
        self._window_size = window.get_size()
        array = self._svg_to_npim(self._svg_source, background_color=background_color)
        pygame.surfarray.blit_array(self._surface, array)
        window.fill(background_color)
        window.blit(self._surface, self.surface_position)
        pygame.display.flip()

    def _surface_to_npim(self, surface: cairosvg.surface.PNGSurface):
        """Transforms a Cairo `surface` into a numpy array."""
        # a copy must be made to avoid a seg fault if the backing array disappears... (not sure why this happens!)
        surface = surface.cairo
        H, W = surface.get_height(), surface.get_width()
        im = np.frombuffer(surface.get_data(), np.uint8)
        im.shape = (H, W, 4)  # for RGBA
        im = im[:, :, :3].transpose(1, 0, 2)[:, :, ::-1].copy()
        return im

    def _svg_to_npim(self, svg_bytestring, dpi=96, background_color="#ffffff"):
        """Renders a svg bytestring as a RGB image in a numpy array."""
        tree = cairosvg.parser.Tree(bytestring=svg_bytestring)
        output_size = self.surface_size
        # this will render to the surface while maintaining the aspect ratio - cool!
        # to compute true window position we need to manaully compute this new position/size of the svg image
        # note that the top-level svg position (x and y attributes) are ignored when rendering, we need to handle this ourselves
        surf = cairosvg.surface.PNGSurface(
            tree,
            None,
            dpi=dpi,
            background_color=background_color,
            output_width=output_size[0],
            output_height=output_size[1],
        )
        # compute the scaling factor this is computed when rendering the svg above
        # we need to manually compute it here to to be able to transform points to
        # svg space and position the canvas surface
        svg_size, sur_size = self.svg_size, self.surface_size
        scaling_factor_width = sur_size[0] / svg_size[0]
        scaling_factor_height = sur_size[1] / svg_size[1]
        # TODO we are going to want to center the image aswell..
        self._scaling_factor = min(scaling_factor_width, scaling_factor_height)

        # the aspect ratio is maintained, but the image is always centered in the output surface,
        # we need to compute this centering manually to be able to reconstruct the svg coordinate transform
        self._surface_offset = (
            (sur_size[0] - (svg_size[0] * self._scaling_factor)) / 2,
            (sur_size[1] - (svg_size[1] * self._scaling_factor)) / 2,
        )
        # surf.finish()
        return self._surface_to_npim(surf)

    def elements_under(
        self, point: tuple[float, float], transform: bool = False
    ) -> list[str]:
        """Gets all svg element `id`s that are under the given `point`.

        Args:
            point (Tuple[float, float]): to check under, expected svg space unless `transform=True` in which case window space (pixels) is expected.
            transform (bool, optional): whether to transform the given `point` to svg space. Defaults to False.

        Returns:
            List[str]: list of element ids that are under the `point`.
        """
        if transform:
            point = self.pixel_to_svg(point)
        return elements_under(self._svg_tree, point)


def point_in_rect(rect, point):
    """Determine if a point is inside a rectangle."""
    try:
        x, y = point
        rx, ry, width, height = (
            float(rect.get(attr)) for attr in ("x", "y", "width", "height")
        )
    except ValueError as e:
        missing = [
            attr for attr in ("x", "y", "width", "height") if attr not in rect.attrib
        ]
        raise ValueError(
            f"Missing required attributes {missing} on rect: '{rect.get('id', '<MISSING ID>')}'"
        ) from e
    return rx <= x <= rx + width and ry <= y <= ry + height, point


def point_in_circle(circle, point):
    """Determine if a point is inside a circle."""
    x, y = point
    cx, cy, r = (float(circle.get(attr)) for attr in ("cx", "cy", "r"))
    return ((x - cx) ** 2 + (y - cy) ** 2) <= r**2, point


def parse_transform(transform: str) -> tuple[float, float, tuple[float, float]]:
    """Parses an svg `transform` attribute extracting the `scale`, `rotation` and `translation`."""
    if transform is None:
        return ((1, 1), None, None)
    scale_match = re.search(r"scale\(([^)]+)\)", transform)
    rotation_match = re.search(r"rotate\(([^)]+)\)", transform)
    translate_match = re.search(r"translate\(([^)]+)\)", transform)

    scale = tuple(map(float, scale_match.group(1).split(","))) if scale_match else None
    rotation = (
        tuple(map(float, rotation_match.group(1).split(",")))
        if rotation_match
        else None
    )
    translate = (
        tuple(map(float, translate_match.group(1).split(",")))
        if translate_match
        else None
    )
    scale = scale if scale else (1.0, 1.0)
    if len(scale) == 1:
        scale = (scale[0], scale[0])
    # TODO rotate
    # if len(rotation) == 1:
    #    return (rotation, 0.0, 0.0)
    # TODO translate
    return scale, rotation, translate


def in_svg(node: ET.ElementBase, point: tuple[float, float]):
    """Checks if the point is in the bounds of the svg element `node` and returns this aswell as the point transformed to the nodes coordinate space."""
    x = node.get("x", None)
    y = node.get("y", None)
    width = node.get("width", None)
    height = node.get("height", None)
    scale, rotation, translate = parse_transform(node.get("transform", None))
    assert rotation is None  # not yet supported
    assert translate is None  # not yet supported
    point = list(point)
    isin = True
    if x is not None:
        point[0] -= float(x)
        point[0] /= scale[0]
        isin &= point[0] >= 0.0
        if width is not None:
            isin &= point[0] <= float(width)
    if y is not None:
        point[1] -= float(y)
        point[1] /= scale[1]
        isin &= point[1] >= 0.0
        if height is not None:
            isin &= point[1] <= float(height)
    return isin, point


def in_group(node: ET.ElementBase, point: tuple[float, float]):
    """Checks if the point is in the bounds of the `group` element and returns this aswell as the point transformed to the nodes coordinate space."""
    assert node.get("transform", None) is None  # not supported
    return True, point


def elements_under(node: ET.ElementBase, point: tuple[float, float]):
    """Gets all svg elements ids that are under the given `point`. Note that not all svg transformation are supported.

    Args:
        node: the SVG element to check.
        point (Tuple[float, float]): to check under

    Returns:
        List[str]: list of element ids that are under the `point`.
    """
    SUPPORTED_SHAPES = {
        "{http://www.w3.org/2000/svg}svg": in_svg,
        "{http://www.w3.org/2000/svg}g": in_group,
        "{http://www.w3.org/2000/svg}rect": point_in_rect,
        "{http://www.w3.org/2000/svg}circle": point_in_circle,
    }
    if node.tag not in SUPPORTED_SHAPES:
        # _LOGGER.warn(f"encountered unsupported shape: %s", node.tag)
        return []
    node_id = node.get("id", None)
    # check if we are
    isin, tpoint = SUPPORTED_SHAPES[node.tag](node, point)
    if isin:
        result = [node_id] if node_id else []
        for child in node:
            result.extend(elements_under(child, tpoint))
        return result
    return []



"""This module defines a simple actuator `AvatarActuator` intended to be used by an `Avatar` that will simply forward any user input to the environment."""

from star_ray.agent import Actuator, attempt

from .event import UserInputEvent

__all__ = ("AvatarActuator",)


class AvatarActuator(Actuator):
    """This simple actuator will automatically foward all user input events (see `star_ray_pygame.UserInputEvent` for a full list) to the environment. This allows other agents to subscribe to receive these events, or for them to be processed by the `Ambient`."""

    @attempt
    def attempt(self, action: UserInputEvent) -> UserInputEvent:
        """Default attempt method, simply forwards any `UserInputEvent`s to the environment (assuming the use of a `star_ray.agent.AgentRouted`).

        Args:
            action (UserInputEvent): to forward

        Returns:
            UserInputEvent: the action (returning the action is default attempt method behaviour).
        """
        return action



"""The `star_ray_pygame` package is an extension of [`star_ray`](https://github.com/dicelab-rhul/star-ray) that implements UI functionality backed by the `pygame` and `star_ray_xml` packages.

`star_ray_pygame` implements a specific type of `Agent` known as an `Avatar` which represents a user (typically human) who is interacting with the environment and the agents within it. The UI backend is `pygame`, and `svg` (Scalable Vector Graphics) are used to render the UI. The UI state completely integrated with the state of the environment allowing other agents to view and potentially modify it. This is a dramatic departure from some common software engineering practices such as the MVC (or similar) design pattern. The UI is NOT decoupled from the program state - in fact, for the most part it IS the program state. This package relies on another extension `star_ray_xml` to manage the state of the UI (as well as any other state variables).

This extension has been developed with the intention of exploring human-agent interactions. Agents are able to view a structured version of the UI and modify it as a means of interacting with the user in the most general way. They are able to effectively render text and graphics via their actions. The (human) user may take similar action (via their peripheral devices - mouse, keyboard, etc.) to communicate back (according to whatever functionality the UI is exposing). Agents are able to listen to these user inputs and make futher decisions based on them.

This package serves as a very flexible foundation for supporting researching into human-agent (or AI) interactions on digital devices by leveraging some existing technologies.

See for example, the [`icua`](https://github.com/dicelab-rhul/icua) project.
"""
# turn of auto-formating, the order of the imports matters!

import os

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"

# ======================================================= #
# NOTE: this must appear before importing other modules! (careful! -> auto-format might move it!)
from .utils import _check_libcairo_install

# this will check for install issues with cairo, its a real pain on windows...
_check_libcairo_install()
# ======================================================= #

from star_ray import Environment  # noqa: E402
from .view import View, WindowConfiguration  # noqa: E402
from .avatar import Avatar  # noqa: E402
from .actuator import AvatarActuator  # noqa: E402
from .ambient import SVGAmbient  # noqa: E402


__all__ = (
    "SVGAmbient",
    "View",
    "Avatar",
    "AvatarActuator",
    "WindowConfiguration",
    "Environment",
)



"""Module contain various utilities used in the `star_ray_pygame` package."""

import os
import sys
from star_ray.utils import _LOGGER as LOGGER


def _check_libcairo_install():
    WINDOW_NAME = "nt"
    ON_WINDOWS = os.name == WINDOW_NAME
    if ON_WINDOWS:
        # try to locate GTK installation and add it to path, this is typically not done by default
        gtkbin = r"C:\Program Files\GTK3-Runtime Win64\bin"
        if os.path.exists(gtkbin):
            os.environ["PATH"] = os.pathsep.join((gtkbin, os.environ["PATH"]))

    def _cairosvg_ok():
        try:
            pass
        except OSError as e:
            return (False, e.args[0])
        return (True, None)

    ok, err = _cairosvg_ok()
    if not ok:
        WINDOWS_INSTALL_MESSAGE = """>> On Windows: Please install the latest GTK-3 runtime from: https://github.com/tschoonj/GTK-for-Windows-Runtime-Environment-Installer/releases"""
        OTHER_INSTALL_MESSAGE = """>> On Linux/MacOS: Contact 'star-ray-pygame' package maintainer or check github issues: https://github.com/dicelab-rhul/star-ray-pygame/issues"""
        ERROR_MESSAGE = f"""'star-ray-pygame' requires a cairo installation which may not be installed automatically on some systems.\nCause: {err}\nPlease ensure the required binaries are installed and accessible via your PATH environment variable.\n\n    {WINDOWS_INSTALL_MESSAGE if ON_WINDOWS else OTHER_INSTALL_MESSAGE}\n    >> See 'star-ray-pygame' README for more information: https://github.com/dicelab-rhul/star-ray-pygame.
        """
        LOGGER.error(ERROR_MESSAGE)
        sys.exit(-1)



"""Tests rendering using the View (without an Avatar)."""

from star_ray_pygame import View, WindowConfiguration
from star_ray.event import (
    WindowCloseEvent,
    MouseButtonEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
)
from lxml import etree

WIDTH, HEIGHT = 640, 480
SVG = f"""
<svg x="100" y="100" width="{WIDTH}" height="{HEIGHT}" xmlns="http://www.w3.org/2000/svg">
  <svg x="50" y="50" width="100" height="100">
        <rect id="myrect" x="0" y="0" width="100" height="200" fill="red"/>
        <circle id="mycircle" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2"/>
  </svg>  

  <g x="200" y="200" width="100" height="100">
        <rect id="myrect" x="0" y="0" width="100" height="200" fill="red"/>
        <circle id="mycircle" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2"/>
  </g>
</svg>"""

SVG = f"""<svg x="0" y="0" width="{WIDTH}" height="{HEIGHT}" xmlns="http://www.w3.org/2000/svg">
    <rect id="myrect" x="0" y="0" width="{WIDTH}" height="{HEIGHT}" fill="yellow"/>
    <rect id="myrect" x="-100" y="0" width="{WIDTH}" height="{HEIGHT}" fill="green"/>

    <rect id="myrect" x="0" y="0" width="100" height="100" fill="blue"/>
    <circle id="mycircle" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2"/>
</svg>"""

window_config = WindowConfiguration(
    width=WIDTH, height=HEIGHT, title="svg test", resizable=True, fullscreen=False
)
view = View(window_config)
view.update(etree.fromstring(SVG))

running = True
while running:
    events = view.get_nowait()
    # time.sleep(0.1)
    for event in events:
        if isinstance(event, WindowCloseEvent):
            running = False
        elif isinstance(event, MouseButtonEvent):
            print(event)
        elif isinstance(event, WindowFocusEvent | WindowMoveEvent | WindowResizeEvent):
            print("window event:", event)
            view.render()
    view.render()



"""The module defines the `SVGAmbient` class which is an extension of `star_ray_xml.XMLAmbient`. This ambient supports pub-sub of XML queries, and contains some useful functionality regarding the handling of SVG data."""

from typing import Any
from star_ray.event import ActiveObservation, ErrorActiveObservation
from star_ray.pubsub._action import Subscribe, Unsubscribe
from star_ray import Agent
from star_ray.utils.type_routing import TypeRouter
from star_ray.pubsub import TypePublisher
from star_ray_xml import XMLAmbient, XMLQuery, Update, Select

from .event import Event, WindowCloseEvent, UserInputEvent
from .utils import LOGGER

__all__ = ("SVGAmbient",)


class SVGAmbient(XMLAmbient):
    """An `Ambient` implementation that provides some useful SVG related functionality, as well as pub-sub support for SVG related queries."""

    DEFAULT_SVG_NAMESPACES = {"svg": "http://www.w3.org/2000/svg"}
    DEFAULT_SVG = """<svg:svg id="root" x="0" y="0" xmlns:svg="http://www.w3.org/2000/svg"></svg:svg>"""
    DEFAULT_SVG_SIZE = (640, 640)
    DEFAULT_SVG_POSITION = (0, 0)

    def __init__(
        self,
        agents: list[Agent],
        svg_size: tuple[float, float] | None = None,
        svg_position: tuple[float, float] | None = None,
        svg_namespaces: dict[str, str] | None = None,
        terminate_on_window_close: bool = True,
        **kwargs,
        # TODO add an option of the root canvas id?
    ):
        """Constructor.

        This will create a root SVG canvas to which all other elements will be drawn. The root svg element id attribute is `root`. This can be changed with an `update` query is desired.

        Args:
            agents (list[Agent]): list of agents to add to this Ambient initially.
            svg_size (tuple[float, float] | None, optional): size of the root svg element. Defaults to None, meaning `width` and `height` will remain undefined.
            svg_position (tuple[float, float] | None, optional): position of the root svg element. Defaults to None, which sets both `x` and `y` to zero.
            svg_namespaces (dict[str, str] | None, optional): the namespaces used. Defaults to the standard svg namespace URI mapping (see `SVGAmbient.DEFAULT_SVG_NAMESPACES`).
            terminate_on_window_close (bool, optional): whether to terminate the environment if a `WindowClosedEvent` occurs, this typically means the user has exited the simulation. Defaults to True.
            kwargs (dict[str,Any]): additional optional arguments.
        """
        # this publisher is used in __subscribe__ to publish events to agents that want them.
        self._publisher = TypePublisher()

        svg = SVGAmbient.DEFAULT_SVG
        svg_namespaces = svg_namespaces if svg_namespaces else dict()
        for k, v in SVGAmbient.DEFAULT_SVG_NAMESPACES.items():
            svg_namespaces.setdefault(k, v)
        super().__init__(agents, xml=svg, namespaces=svg_namespaces, **kwargs)

        self._terminate_on_window_close = terminate_on_window_close
        self._initialise_root(svg_size=svg_size, svg_position=svg_position)

        # this will resolve the types present in `on_user_input_event` type hint and ensure the relevant events are correctly forwarded
        try:
            self._user_input_types = tuple(
                TypeRouter.resolve_first_argument_types(self.on_user_input_event)
            )
            LOGGER.debug(f"user input types: {self._user_input_types}")
        except TypeError as e:
            raise TypeError(
                f"Invalid type hint on method {self.on_user_input_event.__name__}"
            ) from e

    def _initialise_root(
        self,
        svg_size: tuple[float, float] = None,
        svg_position: tuple[float, float] = None,
        **kwargs,
    ):
        """Initialise attributes of the root svg element."""
        # see what has already been defined
        root_attributes = self.__select__(
            Select.new("/svg:svg", ["x", "y", "width", "height"])
        ).values[0]
        # filter out None values
        root_attributes = {k: v for k, v in root_attributes.items() if v is not None}

        if svg_size:
            assert len(svg_size) == 2
            root_attributes["width"] = svg_size[0]
            root_attributes["height"] = svg_size[1]
        if svg_position:
            assert len(svg_position) == 2
            root_attributes["x"] = svg_position[0]
            root_attributes["y"] = svg_position[1]

        root_attributes.setdefault("x", SVGAmbient.DEFAULT_SVG_POSITION[0])
        root_attributes.setdefault("y", SVGAmbient.DEFAULT_SVG_POSITION[1])
        root_attributes.setdefault("width", SVGAmbient.DEFAULT_SVG_SIZE[0])
        root_attributes.setdefault("height", SVGAmbient.DEFAULT_SVG_SIZE[1])
        result = self.__update__(Update.new("/svg:svg", root_attributes))
        if isinstance(result, ErrorActiveObservation):
            raise result.exception()  # something bad happened...

    def on_xml_event(self, action: XMLQuery) -> None:
        """A callback that is triggered immediately AFTER an XMLQuery is executed in `__update__`.

        Args:
            action (XMLQuery): the action
        """
        assert isinstance(action, XMLQuery)

    def on_user_input_event(self, action: UserInputEvent):
        """Callback that is triggered when a user input action is executed in `__update__`. The type hint `UserInputEvent` is used to determine which user input events can be processed by this ambient, it MUST be provided by an override of this method.

        Args:
            action (UserInputEvent): action
        """
        LOGGER.debug("User input event: %s", action)

    def on_unknown_event(self, action: Event):
        """Callback that is triggered when an action of unknown type is executed in `__update__`. This may be overriden for custom handling of such events. Otherwise a ValueError will be raised.

        Args:
            action (Event): unknown action

        Raises:
            ValueError: raised by default.
        """
        # TODO this should be an UnknownEventTypeError? make this a thing
        raise ValueError(f"Action: {action} has unknown type: {type(action)}.")

    def on_exit_event(self, action: WindowCloseEvent):
        """Callback that is triggered when a `WindowCloseEvent` is executed in `__update__`. By default this will flag the environment from termination. This behaviour can be overriden in a subclass.

        Args:
            action (WindowCloseEvent): window close action
        """
        assert isinstance(action, WindowCloseEvent)
        if self._terminate_on_window_close:
            LOGGER.debug("Window closed: %s, shutting down...", action)
            self._is_alive = False  # flag trigger termination

    def __subscribe__(
        self, action: Subscribe | Unsubscribe
    ) -> ActiveObservation | ErrorActiveObservation:
        """Subscribe to receive XMLQuery events from this environment. This will be called automatically when the relevant actions are executed and should not be called manually. Supported actions are only those that inherit from `star_ray_xml.XMLQuery`.

        Args:
            action (Subscribe | Unsubscribe): the subscription action

        Raises:
            TypeError: if the action is not a subscription action

        Returns:
            ActiveObservation | ErrorActiveObservation: the resulting observation - confirmation that the subscription was successful (or not).
        """
        if isinstance(action, Subscribe):
            self._publisher.subscribe(action.topic, action.subscriber)
            return ActiveObservation(action_id=action)
        elif isinstance(action, Unsubscribe):
            self._publisher.unsubscribe(action.topic, action.subscriber)
            return ActiveObservation(action_id=action)
        else:
            raise TypeError(f"Unknown subscription action type: {type(action)}")

    def __update__(self, action: Any) -> ActiveObservation | ErrorActiveObservation:
        """Execute a write action in this `Ambient`.

        Supported action types:
        - Any action that inherits from `XMLQuery`
        - User input actions (types specified by the type hint in `SVGAmbient.on_user_input_event`, see method docs for details.)
        - `WindowCloseEvent` if `terminate_on_window_close` was set to True.

        Unsupported actions will be routed to `SVGAmbient.on_unknown_event`, which may raise an exception.

        To support additional action types override this method (make sure to call `super().__update__`).
        To support additional user input action type override`on_user_input_event` and add new types to the type hints
        - it is best to keep existing type hints unless you explicitly want them to be treated as unknown types.

        Args:
            action (Any): action

        Returns:
            ActiveObservation | ErrorActiveObservation: the resulting observation
        """
        try:
            result = None
            if isinstance(action, XMLQuery):
                result = super().__update__(action)
                self.on_xml_event(action)
            elif isinstance(action, self._user_input_types):
                result = self.on_user_input_event(action)
            else:
                result = self.on_unknown_event(action)
            # exit on close?
            if isinstance(action, WindowCloseEvent):
                self.on_exit_event(action)
            # publish the action for any that wish to see it
            self._publisher.publish(action)
            return result
        except Exception as e:
            return ErrorActiveObservation.from_exception(action, e)



"""Module defining all user input event types and other userful event types such as XML query primitives. Many are collected from `star_ray` and exposed here. Notably the `MouseButtonEvent` and `MouseMotionEvent` have been redefined with additional fields. See these classes for details."""

from star_ray.event import (
    Event,
    MouseButtonEvent as _MouseButtonEvent,
    MouseMotionEvent as _MouseMotionEvent,
    KeyEvent,
    WindowCloseEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowOpenEvent,
    WindowResizeEvent,
    ScreenSizeEvent,
)
from star_ray_xml import XMLQuery, XPathQuery, Update, Select, Insert, Delete, Replace


class MouseButtonEvent(_MouseButtonEvent):
    """An extension of the original `star_ray.event.MouseButtonEvent` which now operates in SVG space by default. The raw (pixel) position is instead stored in the `position_raw` field."""

    position_raw: tuple[float, float] | tuple[int, int]


class MouseMotionEvent(_MouseMotionEvent):
    """An extension of the original `star_ray.event.MouseMotionEvent` which now operates in SVG space by default. The raw (pixel) position is instead stored in the `position_raw` field, and similarly for the relative position in `relative_raw`."""

    position_raw: tuple[float, float] | tuple[int, int]
    relative_raw: tuple[float, float] | tuple[int, int]


MouseEvent = MouseButtonEvent | MouseMotionEvent
WindowEvent = (
    WindowCloseEvent
    | WindowOpenEvent
    | WindowFocusEvent
    | WindowMoveEvent
    | WindowResizeEvent
)
UserInputEvent = MouseEvent | KeyEvent | WindowEvent

__all__ = (
    "Event",
    "UserInputEvent",
    "MouseEvent",
    "WindowEvent",
    "MouseButtonEvent",
    "MouseMotionEvent",
    "KeyEvent",
    "WindowCloseEvent",
    "WindowFocusEvent",
    "WindowMoveEvent",
    "WindowOpenEvent",
    "WindowResizeEvent",
    "ScreenSizeEvent",
    "XMLQuery",
    "XPathQuery",
    "Update",
    "Select",
    "Insert",
    "Delete",
    "Replace",
)



"""Test that runs a simple simulation with an `Avatar`."""

from star_ray_pygame import SVGAmbient, WindowConfiguration
from star_ray import Environment
from star_ray_pygame.avatar import Avatar
from star_ray_pygame.event import MouseButtonEvent
from star_ray_pygame.actuator import AvatarActuator, attempt
from star_ray_pygame.utils import LOGGER
from star_ray_xml import insert, update
import logging

LOGGER.setLevel(logging.DEBUG)

WIDTH, HEIGHT = 640, 480
NAMESPACES = {"svg": "http://www.w3.org/2000/svg"}

RECT1 = f"""<svg:rect id="myrect1" x="0" y="0" width="{WIDTH}" height="{HEIGHT}" fill="yellow" xmlns:svg="http://www.w3.org/2000/svg"/>"""
RECT2 = f"""<svg:rect id="myrect2" x="-100" y="0" width="{WIDTH}" height="{HEIGHT}" fill="green" xmlns:svg="http://www.w3.org/2000/svg"/>"""
RECT3 = """<svg:rect id="myrect3" x="0" y="0" width="100" height="100" fill="blue xmlns:svg="http://www.w3.org/2000/svg""/>"""
CIRCLE1 = """<svg:circle id="mycircle1" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2" xmlns:svg="http://www.w3.org/2000/svg"/>"""

window_config = WindowConfiguration(
    width=WIDTH, height=HEIGHT, title="svg test", resizable=True, fullscreen=False
)


class TestActuator(AvatarActuator):
    """Test Actuator that changes the colour of a circle when clicking left/right mouse."""

    @attempt
    def on_click(self, event: MouseButtonEvent):  # noqa: D102
        color = ["red", "blue", "yellow"][event.button]
        return update(xpath="/svg:svg/svg:circle", attrs={"fill": color})


avatar = Avatar([], [TestActuator()], window_config=window_config)
ambient = SVGAmbient([avatar], svg_position=(0, 0), svg_size=(WIDTH, HEIGHT))
ambient.__update__(insert("/svg:svg", element=RECT1, index=0))
ambient.__update__(insert("/svg:svg", element=RECT2, index=1))
ambient.__update__(insert("/svg:svg", element=RECT3, index=2))
ambient.__update__(insert("/svg:svg", element=CIRCLE1, index=3))

environment = Environment(ambient, wait=0.1)
environment.run()



"""Tests rendering using the View (without an Avatar) with asyncio, see also `test/test_basic_render.py`."""

import asyncio
from lxml import etree
from star_ray.event import (
    WindowCloseEvent,
    MouseButtonEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
)

from star_ray_pygame import View, WindowConfiguration


WIDTH, HEIGHT = 640, 480
SVG = f"""
<svg width="{WIDTH}" height="{HEIGHT}" xmlns="http://www.w3.org/2000/svg">
  <svg x="50" y="50" width="100" height="100">
        <rect id="myrect" x="0" y="0" width="100" height="200" fill="red"/>
        <circle id="mycircle" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2"/>
  </svg>  

  <g x="200" y="200" width="100" height="100">
        <rect id="myrect" x="0" y="0" width="100" height="200" fill="red"/>
        <circle id="mycircle" cx="50" cy="50" r="40" fill="yellow" stroke="black" stroke-width="2"/>
  </g>
</svg>"""

window_config = WindowConfiguration(
    width=WIDTH, height=HEIGHT, title="svg test", resizable=True, fullscreen=False
)


async def main():
    """Main entry point for asyncio."""

    async def task():
        running = True
        view = View(window_config)
        view.update(etree.fromstring(SVG))
        while running:
            events = view.get_nowait()
            await asyncio.sleep(0.1)
            for event in events:
                if isinstance(event, WindowCloseEvent):
                    running = False
                elif isinstance(event, MouseButtonEvent):
                    print(event)
                elif isinstance(
                    event, WindowFocusEvent | WindowMoveEvent | WindowResizeEvent
                ):
                    print("window event:", event)
            view.render()

    task = asyncio.create_task(task())
    await task


asyncio.run(main())



"""Module that defines the `View` class, this class is used to render SVG in a `pygame` window."""

from lxml import etree as ET
from pygame.event import EventType
from typing import Any
import pywinctl as pwc
import pygame
import time
import copy
import math

from star_ray.utils import _LOGGER
from star_ray.event import (
    Event,
    KeyEvent,
    WindowCloseEvent,
    WindowOpenEvent,
    WindowResizeEvent,
    WindowMoveEvent,
    WindowFocusEvent,
    ScreenSizeEvent,
)
from star_ray.ui import WindowConfiguration

from .event import MouseButtonEvent, MouseMotionEvent
from .cairosurface import CairoSVGSurface

# Constant values
PYGAME_KEYDOWN = pygame.KEYDOWN
PYGAME_KEYUP = pygame.KEYUP
PYGAME_QUIT = pygame.QUIT
PYGAME_MOUSEDOWN = pygame.MOUSEBUTTONDOWN
PYGAME_MOUSEUP = pygame.MOUSEBUTTONUP
PYGAME_MOUSEMOTION = pygame.MOUSEMOTION
PYGAME_WINDOWRESIZE = pygame.VIDEORESIZE
PYGAME_WINDOWMOVE = pygame.USEREVENT + 1
PYGAME_WINDOWFOCUS = pygame.USEREVENT + 2
PYGAME_WINDOWOPEN = pygame.USEREVENT + 3
PYGAME_WINDOWCLOSE = PYGAME_QUIT  # Alias for PYGAME_QUIT
PYGAME_SCREENSIZE = pygame.USEREVENT + 4

PYGAME_LEFTMOUSEBUTTON = 1
PYGAME_MIDDLEMOUSEBUTTON = 2
PYGAME_RIGHTMOUSEBUTTON = 3

MOUSE_BUTTON_MAP = {
    PYGAME_LEFTMOUSEBUTTON: MouseButtonEvent.BUTTON_LEFT,
    PYGAME_MIDDLEMOUSEBUTTON: MouseButtonEvent.BUTTON_MIDDLE,
    PYGAME_RIGHTMOUSEBUTTON: MouseButtonEvent.BUTTON_RIGHT,
}


class View:
    """Class that an `Avatar` will use to render SVG in a pygame window and trigger/handle user input events."""

    def __init__(
        self,
        window_config: WindowConfiguration,
        surface_size: tuple[int, int] | None = None,
    ):
        """Constructor.

        Args:
            window_config (WindowConfiguration): configuration of the pygame window.
            surface_size (tuple[int, int], optional): size of the render surface (in pixels). Defaults to None, meaning it will be derived from the supplied window size.
        """
        pygame.init()
        self._window_config = copy.deepcopy(window_config)
        self._screen_size = get_screen_size()

        flags = 0
        if window_config.resizable:
            # this should be false until we can resize/scale the context...
            flags |= pygame.RESIZABLE
        if window_config.fullscreen:
            # set the window size to be the screen size if full screen
            self._window_config.width = self._screen_size[0]
            self._window_config.height = self._screen_size[1]
            flags |= pygame.FULLSCREEN
        self._window_flags = flags
        self._window = pygame.display.set_mode(
            self.window_size, flags=self._window_flags
        )
        if self._window_config.title:
            pygame.display.set_caption(self._window_config.title)
        if surface_size is None:
            PADDING = 20
            surface_size = (
                self.window_size[0] - PADDING,
                self.window_size[1] - PADDING,
            )
            # print(surface_size)
            # print(self.window_size)
        # this is where the svg is rendered. It should be the same size as the svg image
        self._surface = CairoSVGSurface(surface_size)
        self._closed = False
        self._pwc_window = None
        self._pwc_window = self._setup_pwc_window()
        # initial events that are the window position/size
        time.sleep(0.1)  # TODO this is unfortunate...
        window_info = self.get_window_info()
        self._window_resize_callback(window_info["size"])
        self._window_moved_callback(window_info["position"])
        self._screen_size_callback(self._screen_size)

    def get_window_info(self) -> dict[str, Any]:
        """Get information about the display window, including: title, position (on the monitor) and size (in pixels).

        Returns:
            dict[str,Any]: window information
        """
        return dict(
            title=self._pwc_window.title,
            position=self._pwc_window.position,
            size=self._pwc_window.size,
        )

    def get_window_aspect(self) -> float:
        """Get the aspect ratio of the window.

        Returns:
            float: aspect ratio
        """
        return self.window_size[0] / self.window_size[1]

    def get_screen_info(self) -> dict[str, Any]:
        """Get information about the screen (monitor), including the monitor index and size.

        NOTE: Currently only single monitor setups are supported, the monitor index will always be zero.

        Returns:
            dict[str,Any]: screen information
        """
        return dict(monitor=0, size=self._screen_size)

    def update(self, svg_tree: ET.ElementBase):
        """Update the internal svg in preparation for rendering. This will typically preceed a call to `View.render`.

        Args:
            svg_tree (ET.ElementBase): root of the svg tree.
        """
        self._surface.update(svg_tree)

    def render(self) -> None:
        """Render svg to the window. This will typically follow a call to `View.update`."""
        self._surface.render(self._window)

    def elements_under(
        self, point: tuple[float, float], transform: bool = False
    ) -> list[str]:
        """Gets all svg element `id`s that are under the given `point`.

        Args:
            point (Tuple[float, float]): to check under, expected svg space unless `transform=True` in which case window space (pixels) is expected.
            transform (bool, optional): whether to transform the given `point` to svg space. Defaults to False.

        Returns:
            List[str]: list of element ids that are under the `point`.
        """
        return self._surface.elements_under(point, transform=transform)

    def pixel_to_svg(self, point: tuple[float, float]) -> tuple[float, float]:
        """Transforms a point from pixel space to svg space.

        Args:
            point (tuple[float, float]): to transform

        Returns:
            tuple[float, float]: transformed point
        """
        return self._surface.pixel_to_svg(point)

    def svg_to_pixel(self, point: tuple[float, float]) -> tuple[float, float]:
        """Transform a point from svg space to pixel space.

        Args:
            point (tuple[float, float]): to transform

        Returns:
            tuple[float, float]: transformed point
        """
        return self._surface.svg_to_pixel(point)

    def pixel_scale_to_svg_scale(
        self, point: tuple[float, float]
    ) -> tuple[float, float]:
        """Scale a point (typically a (width, height) pair or similar) that lies in pixel space to svg space. Only scaling is applied.

        Args:
            point (tuple[float, float]): point to scale

        Returns:
            tuple[float, float]: scaled point
        """
        return self._surface.pixel_scale_to_svg_scale(point)

    def svg_scale_to_pixel_scale(
        self, point: tuple[float, float]
    ) -> tuple[float, float]:
        """Scale a point (typically a (width, height) pair or similar) that lies in svg space to pixel space. Only scaling is applied.

        Args:
            point (tuple[float, float]): point to scale

        Returns:
            tuple[float, float]: scaled point
        """
        return self._surface.svg_scale_to_pixel_scale(point)

    def get_nowait(self) -> list[Event]:
        """Get all user input events that have happened since the last call to this method. Internally this will poll the `pygame` event loop.

        Returns:
            list[Event]: user input events.
        """
        events = []
        for pg_event in pygame.event.get():
            fun = EVENT_MAP.get(pg_event.type, None)
            if fun:
                try:
                    events.append(fun(self, pg_event))
                except Exception as e:
                    _LOGGER.exception(
                        "Failed to convert `pygame` event: %s  to `star_ray` event as an error occured: %s",
                        pg_event,
                        e,
                    )
        return events

    @property
    def is_open(self) -> bool:
        """Is the window currently open?

        Returns:
            bool: True if the window is open, False otherwise.
        """
        return not self._closed

    def close(self) -> None:
        """Close the UI - terminate pygame and clean up."""
        self._closed = True
        self._pwc_window.watchdog.stop()
        pygame.quit()

    @property
    def window_config(self) -> WindowConfiguration:
        """Get current window configuration. This is a read only quantity, modifying it will have no effect, if you want to change any window property use the relevant `View` attributes.

        Returns:
            WindowConfiguration: current window configuration.
        """
        return copy.deepcopy(self._window_config)  # read only!

    @property
    def window_size(self) -> tuple[int, int]:
        """Getter for the current window size.

        Returns:
            tuple[int, int]: current window size
        """
        return (int(self._window_config.width), int(self._window_config.height))

    @window_size.setter
    def window_size(self, value: tuple[int, int]):
        """Setter for the window size. This operation is not cheap, it is not advisable to resize the window frequently.

        Args:
            value (tuple[int,int]): new window size.
        """
        window_size = (self._window_config.width, self._window_config.height)
        if window_size != value:
            # NOTE: resizing using set_mode will break pywinctl watch dog because a new window is created
            # be careful using pygame.display.set_mode in this class!
            width, height = int(math.ceil(value[0])), int(math.ceil(value[1]))
            self._window = pygame.display.set_mode(
                (width, height), flags=self._window_flags
            )
            # pygame creates a new window... pwc needs to find it again
            self._pwc_window = self._setup_pwc_window()
            self._window_config.width = width
            self._window_config.height = height
            # this doesnt seem to get called during `set_mode` above...
            self._window_resize_callback((width, height))

    def _window_open_callback(self):
        """Callback for when the pygame window is opened for the first time. A custom pygame event is added to the queue internally. The associated event timestamp is not accurate, but the event serves as an indication of when this view is ready to for updating/rendering."""
        pygame.event.post(pygame.event.Event(PYGAME_WINDOWOPEN))

    def _window_focus_callback(self, has_focus):
        """Callback for `pywinctl` when the pygame window losses or gains focus. A custom pygame event is added to the queue internally."""
        pygame.event.post(pygame.event.Event(PYGAME_WINDOWFOCUS, has_focus=has_focus))

    def _window_moved_callback(self, position):
        """Callback for `pywinctl` when the pygame window is moved. A custom pygame event is added to the queue internally."""
        pygame.event.post(pygame.event.Event(PYGAME_WINDOWMOVE, position=position))

    def _window_resize_callback(self, size):
        """Callback for `pywinctl` when the pygame window is resized. A custom pygame event is added to the queue internally."""
        pygame.event.post(pygame.event.Event(PYGAME_WINDOWRESIZE, size=size))

    def _screen_size_callback(self, size):
        """Callback for when the screen size is discovered for the first time. A custom pygame event is added to the queue internally."""
        pygame.event.post(pygame.event.Event(PYGAME_SCREENSIZE, size=size))

    def _setup_pwc_window(self) -> pwc.Window:
        """Getter for the `pywinctl` window object. This object is used to watch changes to the pygame window and addresses some pygame limitations."""
        if self._pwc_window:
            self._pwc_window.watchdog.stop()

        title = self._window_config.title
        windows = pwc.getWindowsWithTitle(self._window_config.title)
        if len(windows) == 0:
            raise ValueError(
                f"Failed to get screen info: couldn't find pygame window with title: {title}"
            )
        if len(windows) > 1:
            raise ValueError(
                f"Failed to get screen info: found multiple windows ({len(windows)}) with the same title: {title}"
            )
        pwc_window = windows[0]
        pwc_window.watchdog.start(
            isActiveCB=self._window_focus_callback,
            movedCB=self._window_moved_callback,
            interval=0.05,
        )
        return pwc_window


def get_screen_size():
    """Getter for the screen size. This will typically not change during the course of the simulation."""
    screen_sizes = pygame.display.get_desktop_sizes()
    assert len(screen_sizes) == 1  # TODO multiple monitors not supported...
    return screen_sizes[0]


def create_window_move_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> WindowMoveEvent:
    """Creates a `WindowMoveEvent` from a `pygame` window move event."""
    if pygame_event.type != PYGAME_WINDOWMOVE:
        raise ValueError(
            "The provided pygame event is not a `PYGAME_WINDOWMOVE` event."
        )
    return WindowMoveEvent(position=pygame_event.position)


def create_window_focus_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> WindowFocusEvent:
    """Creates a `WindowFocusEvent` from a `pygame` window focus event."""
    if pygame_event.type != PYGAME_WINDOWFOCUS:
        raise ValueError(
            "The provided pygame event is not a `PYGAME_WINDOWFOCUS` event."
        )
    return WindowFocusEvent(has_focus=pygame_event.has_focus)


def create_window_resize_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> WindowResizeEvent:
    """Creates a `WindowResizeEvent` from a `pygame` window resize event."""
    if pygame_event.type != PYGAME_WINDOWRESIZE:
        raise ValueError(
            "The provided pygame event is not a `PYGAME_WINDOWRESIZE` event."
        )
    return WindowResizeEvent(size=tuple(pygame_event.size))


def create_screen_size_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> ScreenSizeEvent:
    """Creates a `ScreenSizeEvent` from a `pygame` screen size event."""
    if pygame_event.type != PYGAME_SCREENSIZE:
        raise ValueError(
            "The provided pygame event is not a `PYGAME_SCREENSIZE` event."
        )
    return ScreenSizeEvent(size=tuple(pygame_event.size))


def create_window_close_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> WindowCloseEvent:
    """Creates an `WindowCloseEvent` from a `pygame` window close event."""
    if pygame_event.type != PYGAME_QUIT:
        raise ValueError("The provided pygame event is not a `PYGAME_QUIT` event.")
    return WindowCloseEvent()


def create_window_open_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> WindowOpenEvent:
    """Creates an `WindowOpenEvent` from a `pygame` window open event."""
    if pygame_event.type != PYGAME_WINDOWOPEN:
        raise ValueError(
            "The provided pygame event is not a `PYGAME_WINDOWOPEN` event."
        )
    return WindowOpenEvent()


def create_key_event_from_pygame_event(view: View, pygame_event: EventType) -> KeyEvent:
    """Creates a `KeyEvent` instance from a `pygame` keyboard event."""
    if pygame_event.type not in (PYGAME_KEYDOWN, PYGAME_KEYUP):
        raise ValueError(
            "The provided pygame event is not a `KEYDOWN` or `KEYUP` event."
        )
    status = KeyEvent.DOWN if pygame_event.type == PYGAME_KEYDOWN else KeyEvent.UP
    return KeyEvent(
        key=pygame.key.name(pygame_event.key), keycode=pygame_event.key, status=status
    )


def create_mouse_button_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> MouseButtonEvent:
    """Creates a `MouseButtonEvent` instance from a `pygame` mouse event."""
    if pygame_event.type not in (PYGAME_MOUSEDOWN, PYGAME_MOUSEUP):
        raise ValueError(
            "The provided pygame event is not a `MOUSEDOWN` or `MOUSEUP` event."
        )
    status = (
        MouseButtonEvent.DOWN
        if pygame_event.type == PYGAME_MOUSEDOWN
        else MouseButtonEvent.UP
    )
    position_raw = pygame_event.pos
    position = view.pixel_to_svg(position_raw)
    target = view.elements_under(position, transform=False)

    return MouseButtonEvent(
        button=MOUSE_BUTTON_MAP[pygame_event.button],
        position=position,
        status=status,
        target=target,
        position_raw=position_raw,
    )


def create_mouse_motion_event_from_pygame_event(
    view: View,
    pygame_event: EventType,
) -> MouseMotionEvent:
    """Creates a `MouseMotionEvent` instance from a `pygame` mouse movement event."""
    if pygame_event.type != PYGAME_MOUSEMOTION:
        raise ValueError("The provided pygame event is not a `MOUSEMOTION` event.")

    position_raw = pygame_event.pos
    relative_raw = pygame_event.rel
    position = view.pixel_to_svg(position_raw)
    relative = view.pixel_scale_to_svg_scale(relative_raw)
    target = view.elements_under(position, transform=False)
    return MouseMotionEvent(
        position=position,
        relative=relative,
        position_raw=position_raw,
        relative_raw=relative_raw,
        target=target,
    )


# Dictionary mapping pygame event types to corresponding event creation functions
EVENT_MAP = {
    PYGAME_KEYDOWN: create_key_event_from_pygame_event,
    PYGAME_KEYUP: create_key_event_from_pygame_event,
    PYGAME_QUIT: create_window_close_event_from_pygame_event,
    PYGAME_MOUSEDOWN: create_mouse_button_event_from_pygame_event,
    PYGAME_MOUSEUP: create_mouse_button_event_from_pygame_event,
    PYGAME_MOUSEMOTION: create_mouse_motion_event_from_pygame_event,
    PYGAME_WINDOWRESIZE: create_window_resize_event_from_pygame_event,
    PYGAME_WINDOWMOVE: create_window_move_event_from_pygame_event,
    PYGAME_WINDOWFOCUS: create_window_focus_event_from_pygame_event,
    PYGAME_SCREENSIZE: create_screen_size_event_from_pygame_event,
}



"""Module defines the `Avatar` class, see class for details."""

from star_ray import Sensor, Actuator, Component
from star_ray.event import ActiveObservation, ErrorObservation, Event
from star_ray_xml import _XMLState, XMLSensor, XMLQuery
from star_ray.ui import WindowConfiguration
from star_ray.agent import AgentRouted, IOSensor, observe
from .view import View, get_screen_size


class Avatar(AgentRouted):
    """Class implementing an `Agent` that acts on behalf of a (human) user. It displays a UI that the user can interact with. User input (clicks, key presses, etc.) will be captured and executed via this agents actuators. Such actuators will typically convert these user input events to actions that will mutate the environment in some way, or request new information to be displayed in the UI. It is also typical for user input events to be made avaliable to other agents in the environment (via `star_ray`s pub-sub mechanism), this can be done simply by forwarding them directly to the environment via a dedicated actuator."""

    def __init__(
        self,
        sensors: list[Sensor] = None,
        actuators: list[Actuator] = None,
        window_config: WindowConfiguration | None = None,
        **kwargs,
    ):
        """Constructor.

        Args:
            sensors (list[Sensor], optional): list of sensors. Defaults to None. An `XMLSensor` will always be added to this list.
            actuators (list[Actuator], optional): list of actuators. Defaults to None.
            window_config (WindowConfiguration | None, optional): window configuration for the pygame window. Defaults to None, see default in `star_ray_pygame.View`.
            kwargs (dict[str,Any]): additional optional arguments.
        """
        actuators = actuators if actuators else []
        sensors = sensors if sensors else []
        sensors.append(XMLSensor())
        if window_config is None:
            window_config = WindowConfiguration()  # use default values
        self._view = View(window_config=window_config)
        # this sensor will get all relevant user input events!
        sensors.append(IOSensor(self._view))
        self._state = None  # set on the first cycle when svg data has been sensed
        super().__init__(sensors, actuators, **kwargs)

    @property
    def xml_sensor(self) -> XMLSensor:
        """Getter for the `XMLSensor` that is attached to this agent.

        Returns:
            XMLSensor: the sensor
        """
        return next(iter(self.get_sensors(oftype=XMLSensor)), None)

    def get_screen_info(self):
        """Get information about the screen (monitor), including the monitor index and size.

        NOTE: Currently only single monitor setups are supported, the monitor index will always be zero.

        Returns:
            dict[str,Any]: screen information
        """
        return self._view.get_screen_info()

    def get_window_info(self):
        """Get information about the display window, including: title, position (on the monitor) and size (in pixels).

        Returns:
            dict[str,Any]: window information
        """
        return self._view.get_window_info()

    @observe
    def on_error_observation(self, observation: ErrorObservation):
        """Callback for error observations, by default an exception is raised. Override for custom handling of these errors.

        Args:
            observation (ErrorObservation): observation containing the error.

        Raises:
            observation.exception: the exception contained in the observation (raised by default).
        """
        raise observation.exception()

    @observe
    def on_observation(self, observation: ActiveObservation, component: Component):
        """Callback when receiving an observation. By default this is only expected to be called upon the first observation made by the attached `XMLSensor` which senses the full state of the environment (all svg code).

        Args:
            observation (ActiveObservation): observation containing svg code for rendering.
            component (Component): component that the observation originated from
        """
        if isinstance(component, XMLSensor):
            if observation.values:
                # TODO we can do some checks to determine whether this is infact the first observation...
                # otherwise, this should only called on the first cycle to set the initialise xml state
                assert self._state is None
                assert len(observation.values) == 1
                self._state = _XMLState(observation.values[0])
                # TODO this is a bit of a hack, it assumes all namespaces are defined in the root, which may not be the case.
                # TODO make an official note of this, or create an action that can get the actual namespaces from the environment...?
                self._state._namespaces = self._state.get_root().nsmap
            else:
                pass  # this is probably the result of a subscription action

    @observe
    def _on_xml_change(self, observation: XMLQuery):
        """Called whenever an `XMLQuery` event is received via the `XMLSensor` attached to this agent, it will update the state of the view."""
        # update view state from xml query
        observation.__execute__(self._state)

    @observe
    def on_event(self, observation: Event):
        """Automatically route all events that may come from sensors to relevant actuators.

        Args:
            observation (Event): observations to route
        """
        self.attempt(observation)

    def __cycle__(self):  # noqa: D105
        super().__cycle__()
        self.render()

    def render(self) -> None:
        """Render the UI."""
        # TODO consider refactoring the view to make use of XMLState rather than the underlying lxml <- this is a big job...
        self._view.update(self._state.get_root()._base)
        self._view.render()

    @staticmethod
    def get_screen_size() -> tuple[int, int]:
        """Get the screen/monitor size (in pixels).

        Returns:
            tuple[int, int]: screen size
        """
        return get_screen_size()
